!-------------------------------------------------------------------------------

!VERS

! This file is part of Code_Saturne, a general-purpose CFD tool.
!
! Copyright (C) 1998-2013 EDF S.A.
!
! This program is free software; you can redistribute it and/or modify it under
! the terms of the GNU General Public License as published by the Free Software
! Foundation; either version 2 of the License, or (at your option) any later
! version.
!
! This program is distributed in the hope that it will be useful, but WITHOUT
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
! FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
! details.
!
! You should have received a copy of the GNU General Public License along with
! this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
! Street, Fifth Floor, Boston, MA 02110-1301, USA.

!-------------------------------------------------------------------------------

!===============================================================================
! Purpose:
! -------

!> \file cs_user_extra_operations.f90
!>
!> \brief This function is called at the end of each time step, and has a very
!>  general purpose
!>  (i.e. anything that does not have another dedicated user subroutine)
!>
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! Arguments
!______________________________________________________________________________.
!  mode           name          role                                           !
!______________________________________________________________________________!
!> \param[in]     nvar          total number of variables
!> \param[in]     nscal         total number of scalars
!> \param[in]     dt            time step (per cell)
!_______________________________________________________________________________

subroutine cs_f_user_extra_operations &
 ( nvar   , nscal  ,                                              &
   dt     )

!===============================================================================

!===============================================================================
! Module files
!===============================================================================

use paramx
use dimens, only: ndimfb
use pointe
use numvar
use optcal
use cstphy
use cstnum
use entsor
use lagpar
use lagran
use lagdim
use parall
use period
use ppppar
use ppthch
use ppincl
use mesh
use field
use field_operator
use cs_c_bindings
use user_module

!===============================================================================

implicit none

! Arguments

integer          nvar   , nscal

double precision dt(ncelet)

! Local variables

character*80     fname

integer          iel, impout, ivar, isou
integer          nitmx, ii, nptx
integer          iwarn, isweep, iswmax
integer          irangv, f_id

logical          cved

double precision eps, vol
double precision gama, rg, rho0
double precision xmin, smin
double precision machi, machia, machib, aire
double precision mache1, mache2, pe1, pe2, p2
double precision mache, pe
double precision secstar, secstar1, secstar2, secshock
double precision massflow
double precision p0left, p0right, machleft, machright
double precision rho0left, rho0right, t0left, t0right
double precision machtmp, prestmp, rhotmp, temptmp, soundtmp
double precision rvoid
double precision slope, orig, radius2, xyz(3)

double precision, allocatable, dimension(:) :: sec, xsec, varres, w1, w2
double precision, allocatable, dimension(:) :: mach, pres, rho, temp, sound, vel

!===============================================================================

type(solving_info) sinfo

!===============================================================================
! 1. Convergence test
!===============================================================================

! Allocate a temporary array for cells or interior/boundary faces selection
allocate(varres(nvar))

cved = .true.

do ivar = 1, nvar
  f_id = ivarfl(ivar)
  call field_get_key_struct_solving_info(f_id, sinfo)
  cved = cved.and.(sinfo%l2residual.le.1.d-9)
  varres(ivar) = sinfo%l2residual
enddo

! Warning: L2 time residual is computed AFTER extra_operations
cved = cved.and.(ntcabs-ntpabs.gt.1).and.ntcabs.lt.ntmabs

if (cved) then
  ntmabs = ntcabs+1
  write(*,*)'converged at ',ntmabs
end if

if (irangp.le.0.and.ntcabs-ntpabs.eq.2) then
  open(unit=1, file='residuals.dat')
  write(1,'(a34)',advance="no") "#    Time step          Time      "
  do ivar = 1, nvar
    if (ivar.ne.iv.and.ivar.ne.iw) then
      call field_get_name(ivarfl(ivar), fname)
      write(1,'(10a14)',advance="no") fname
    endif
  enddo
  close(1)
endif

if (     irangp.le.0 &
    .and.(ntcabs.eq.2.or.mod(ntcabs,ntlist).eq.0.or.ntcabs.eq.ntmabs)) then
  open(1, file='residuals.dat', position='append')
  write(1,'(i14,10e14.5)',advance="no") ntcabs, ttcabs
  do ivar = 1, nvar
    if (ivar.ne.iv.and.ivar.ne.iw) then
      write(1,'(10e14.5)',advance="no") varres(ivar)
    endif
  enddo
  close(1)
endif

deallocate(varres)

!===============================================================================
! Compute 1D solution at last time step
!===============================================================================

if (ntcabs.eq.ntmabs) then

  ! nptx = number of points in x direction
  ! nitmx = maximum iteration number for the Newton algorithm

  nptx = 100
  nitmx = 10000

  allocate(sec(nptx))
  allocate(xsec(nptx))
  allocate(pres(nptx))
  allocate(mach(nptx))
  allocate(rho(nptx))
  allocate(temp(nptx))
  allocate(sound(nptx))
  allocate(vel(nptx))

  gama   = 1.d0 + cs_physical_constants_r/(xmasmr*cp0-cs_physical_constants_r)
  rg     = cs_physical_constants_r/xmasmr

  ! pe = Static pressure at the exhaust (outlet)

  ! SHOCK
  if (testcase.eq.1) pe = 0.75d5
  ! SUBSONIC
  if (testcase.eq.2) pe = 0.95d5
  ! SUPERSONIC
  if (testcase.eq.3) pe = 0.1d5

  rho0 = p0/(rg*t0)
  write(nfecra,*) '  ** COMPUTATION OF 1D SOLUTION OF A FLOW IN A LAVAL NOZZLE'
  write(nfecra,*) '  ---------------------------------------------------------'
  write(nfecra,*) '--------------------------------------------------------------'
  write(nfecra,*) '  Upstream stagnation conditions of the nozzle'
  write(nfecra,*) '  p0   (Pa)    : ', p0
  write(nfecra,*) '  T0   (K)     : ', t0
  write(nfecra,*) '  rho0 (kg/m3) : ', rho0
  write(nfecra,*) '--------------------------------------------------------------'

  ! Retrieve mesh nodes
  do ii = 1, nptx
    xyz(1) = float(ii-1)/float(nptx-1)*0.32d0
    xyz(2) = 0.d0
    xyz(3) = 0.d0

    call findpt(ncelet, ncel, xyzcen, xyz(1), xyz(2), xyz(3), iel, irangv)

    if (irangp.eq.irangv) then
      xsec(ii) = xyzcen(1,iel)
    endif

    if (irangp.ge.0) then
      call parall_bcast_r(irangv,xsec(ii))
    endif
  enddo

  if (irangp.le.0) then

    impout = impusr(2)

    xmin = 0.1446d0
    do ii = 1, nptx
      if (xsec(ii).le.0.0265d0) then
        sec(ii) = 0.06d0
      elseif (xsec(ii).gt.0.0265d0 .and. xsec(ii).le.0.0565d0) then
        radius2 = 0.06d0**2
        sec(ii) = sqrt(radius2-(xsec(ii)-0.0265d0)**2.d0)
      elseif (xsec(ii).gt.0.0565d0 .and. xsec(ii).le.0.0996d0) then
        slope = (0.027d0-0.0519d0)/(0.0996d0-0.0565d0)
        orig  = 0.027d0-slope*0.0996d0
        sec(ii) = slope*xsec(ii) + orig
      elseif (xsec(ii).gt.0.0996d0 .and. xsec(ii).le.0.1525d0) then
        radius2 = (0.105d0-0.027d0)**2.d0+(0.1446d0-0.0996d0)**2.d0
        sec(ii) = 0.105d0-sqrt(radius2-(xsec(ii)-0.1446d0)**2.d0)
        smin    = 0.105d0-sqrt(radius2-(xmin-0.1446d0)**2.d0)
      else
        slope = (0.03d0-0.0153d0)/(0.32d0-0.1525d0)
        orig  = 0.0153d0-slope*0.1525d0
        sec(ii) = slope*xsec(ii) + orig
      endif
    enddo
    do ii = 1, nptx
      xsec(ii) = xsec(ii) - xmin
      sec(ii)  = 2.d0*sec(ii)
    enddo
    smin = 2.d0*smin

    ! Step 1 : Calculation of the two exhaust pressure separating the three regimes

    ! Calculation of the exhaust Mach number thanks to the areas law
    aire    = sec(nptx)/smin
    machi   = 0.1d0
    iwarn  = 0
    call newton(nitmx,iwarn,gama,machi,aire,mache1)
    machi   = 1.5d0
    call newton(nitmx,iwarn,gama,machi,aire,mache2)

    ! Calculation of the pressure pe1
    call pres_isent(gama,mache1,pe1)
    pe1 = 1.0d0/pe1
    pe1 = p0*pe1

    ! Calculation of the pressure pe2
    call pres_isent(gama,mache2,p2)
    p2 = 1.d0/p2
    p2 = p0*p2
    call pres_shock(gama,mache2,pe2)
    pe2=pe2*p2

    write(nfecra,*) '  Exhaust Mach number and pressure separating the three different'
    write(nfecra,*) '  regimes of the nozzle'
    write(nfecra,*) '  me1      : ', mache1
    write(nfecra,*) '  pe1 (Pa) : ', pe1
    write(nfecra,*) '  me2      : ', mache2
    write(nfecra,*) '  pe2 (Pa) : ', pe2
    write(nfecra,*) '--------------------------------------------------------------'

    ! Step 2 : Calculation of the solution : diverging subsonic nozzle

    if (pe.ge.pe1) then
      write(nfecra,*) '  With pe (Pa) : ', pe
      write(nfecra,*) '  the diverging nozzle is subsonic'
      write(nfecra,*) '--------------------------------------------------------------'

      ! Calculation of the exhaust Mach number
      pe = p0/pe
      call invpres_isent(gama,pe,mache)
      pe = p0/pe

      ! Calculation of the critical section
      call aeras_law(gama,mache,aire,rvoid)
      secstar = sec(nptx)/aire
      write(nfecra,*) '  Critical section (m2) : ', secstar
      write(nfecra,*) '--------------------------------------------------------------'

      ! Calculation of the solution for each x position
      open(impout,file='nozzle_profile.dat')
      write(impout,'(A)') '# x(m) S(m2) Mach P(Pa) rho(kg/m3) T(K) c(m2/s) u(m/s)'
      iwarn = 0
      machi  = 0.1d0
      do ii = 1, nptx
        aire   = sec(ii)/secstar
        call newton(nitmx,iwarn,gama,machi,aire,machtmp)
        call pres_isent(gama,machtmp,prestmp)
        call rho_isent(gama,machtmp,rhotmp)
        call temp_isent(gama,machtmp,temptmp)
        call sound_velocity(gama,p0/prestmp,rho0/rhotmp,soundtmp)
        mach(ii) = machtmp
        pres(ii) = p0/prestmp
        rho(ii)  = rho0/rhotmp
        temp(ii) = t0/temptmp
        sound(ii) = soundtmp
        vel(ii) = machtmp*soundtmp
        write(impout,'(8e14.5)') xsec(ii)+xmin, sec(ii), mach(ii), pres(ii), rho(ii), &
             temp(ii), sound(ii), vel(ii)
      enddo
      close(impout)

      ! Calculation of the mass flow rate,
      ! and writing of the inlet/Exhaust conditions
      massflow = rho(nptx)*mach(nptx)*sound(nptx)*sec(nptx)
      write(nfecra,*) '  Mass flow rate(kg/s) : ', massflow
      write(nfecra,*) '--------------------------------------------------------------'
      write(nfecra,*) '  Inlet conditions'
      write(nfecra,*) '  ui   (m/s)   : ', vel(1)
      write(nfecra,*) '  Mi           : ', mach(1)
      write(nfecra,*) '  pi   (Pa)    : ', pres(1)
      write(nfecra,*) '  rhoi (kg/m3) : ', rho(1)
      write(nfecra,*) '  Ti   (K)     : ', temp(1)
      write(nfecra,*) '--------------------------------------------------------------'
      write(nfecra,*) '  Exhaust conditions'
      write(nfecra,*) '  ue   (m/s)   : ', vel(nptx)
      write(nfecra,*) '  Me           : ', mach(nptx)
      write(nfecra,*) '  pe   (Pa)    : ', pres(nptx)
      write(nfecra,*) '  rhoe (kg/m3) : ', rho(nptx)
      write(nfecra,*) '  Te   (K)     : ', temp(nptx)
      write(nfecra,*) '---------------------------------------------------------------'
      write(nfecra,*) 'Solution profiles are given in nozzle_profile.dat in this order'
      write(nfecra,*) '# x(m) S(m2) Mach P(Pa) rho(kg/m3) T(K) c(m2/s) u(m/s)'


      !===============================================================================
      ! Step 3 :  Calculation of the solution : diverging supersonic nozzle
      !===============================================================================

    elseif (pe.lt.pe2) then

      write(nfecra,*) '  With pe(Pa) : ', pe
      write(nfecra,*) '  the diverging nozzle is supersonic'
      write(nfecra,*) '--------------------------------------------------------------'
      secstar = smin
      write(nfecra,*) '  Critical section (m2) : ', secstar, '(throat)'

      ! Calculation of the solution for each x position
      open(impout,file='nozzle_profile.dat')
      write(impout,'(A)') '# x(m) S(m2) Mach P(Pa) rho(kg/m3) T(K) c(m2/s) u(m/s)'
      iwarn = 0
      do ii = 1, nptx
        if (xsec(ii).lt.0.d0) then
          machi  = 0.1d0
        else
          machi  = 1.1d0
        endif
        aire   = sec(ii)/secstar
        call newton(nitmx,iwarn,gama,machi,aire,machtmp)
        call pres_isent(gama,machtmp,prestmp)
        call rho_isent(gama,machtmp,rhotmp)
        call temp_isent(gama,machtmp,temptmp)
        call sound_velocity(gama,p0/prestmp,rho0/rhotmp,soundtmp)
        mach(ii) = machtmp
        pres(ii) = p0/prestmp
        rho(ii)  = rho0/rhotmp
        temp(ii) = t0/temptmp
        sound(ii) = soundtmp
        vel(ii) = machtmp*soundtmp
        write(impout,'(8e14.5)') xsec(ii)+xmin, sec(ii), mach(ii), pres(ii), rho(ii), &
             temp(ii), sound(ii), vel(ii)
      enddo
      close(impout)

      ! Calculation of the mass flow rate,
      ! and writing of the inlet/Exhaust conditions
      massflow = rho(nptx)*mach(nptx)*sound(nptx)*sec(nptx)
      write(nfecra,*) '  Mass flow rate (kg/s) : ', massflow
      write(nfecra,*) '--------------------------------------------------------------'
      write(nfecra,*) '  Inlet conditions'
      write(nfecra,*) '  ui   (m/s)   : ', vel(1)
      write(nfecra,*) '  Mi           : ', mach(1)
      write(nfecra,*) '  pi   (Pa)    : ', pres(1)
      write(nfecra,*) '  rhoi (kg/m3) : ', rho(1)
      write(nfecra,*) '  Ti   (K)     : ', temp(1)
      write(nfecra,*) '--------------------------------------------------------------'
      write(nfecra,*) '  Exhaust conditions'
      write(nfecra,*) '  ue   (m/s)   : ', vel(nptx)
      write(nfecra,*) '  Me           : ', mach(nptx)
      write(nfecra,*) '  pe   (Pa)    : ', pres(nptx)
      write(nfecra,*) '  rhoe (kg/m3) : ', rho(nptx)
      write(nfecra,*) '  Te   (K)     : ', temp(nptx)
      write(nfecra,*) '---------------------------------------------------------------'
      write(nfecra,*) 'Solution profiles are given in nozzle_profile.dat in this order:'
      write(nfecra,*) '# x(m) S(m2) Mach P(Pa) rho(kg/m3) T(K) c(m2/s) u(m/s)'

      !===============================================================================
      ! Step 4 :  Calculation of the solution : shock in the diverging nozzle
      !===============================================================================

    else

      write(nfecra,*) '  With pe (Pa) : ', pe
      write(nfecra,*) '  there is a schock in the diverging noozle'
      write(nfecra,*) '--------------------------------------------------------------'
      secstar1 = smin
      p0left   = p0
      t0left   = t0
      rho0left = rho0

      ! Iterative process to calculate the downstream stagnation pressure
      prestmp = pe
      p0right = 1.d12
      eps = 1.d-15
      iswmax = 100
      isweep = 1
      iwarn = 0
      machi  = 0.1d0

      do while (isweep.le.iswmax.and.abs(p0right-prestmp).ge.eps)
        p0right = prestmp
        aire   = (sec(nptx)/secstar1)*(p0right/p0left)
        call newton(nitmx,iwarn,gama,machi,aire,machtmp)
        call pres_isent(gama,machtmp,prestmp)
        prestmp = prestmp*pe
        if (iwarn.eq.1) then
          write(nfecra,*) '  iteration = ', isweep, 'p0right (Pa) = ', prestmp, &
               'residu = ', abs(p0right-prestmp)
        endif
        isweep = isweep + 1
      enddo
      write(nfecra,*) ''
      write(nfecra,*) '  dowstream stagnation pressure (Pa) : ', prestmp

      ! Calculation of the Mach number upstream and downstream the shock
      p0right   = prestmp
      t0right   = t0
      rho0right = p0right/(rg*t0right)
      iwarn    = 0
      machia    = 1.0d0
      machib    = 5.0d0
      call dichotomy(iwarn, gama, machia, machib, p0right/p0left, machleft)
      write(nfecra,*) '  Mach number upstream the shock   : ', machleft
      call mach_shock(gama,machleft,machright)
      write(nfecra,*) '  Mach number downstream the shock : ', machright
      write(nfecra,*) '--------------------------------------------------------------'

      ! Calculation of the shock section
      secstar2 = (p0left/p0right)*secstar1
      call aeras_law(gama,machright,secshock,rvoid)
      secshock = secshock*secstar2
      write(nfecra,*) '  Shock section (m2) : ', secshock, 'such that:'
      write(nfecra,*) '  S(throat)=',smin,' < S(shock)=', secshock, &
           ' < S(exhaust)=', sec(nptx)

      ! Calculation of the solution for each x position
      open(impout,file='nozzle_profile.dat')
      write(impout,'(A)') '# x(m) S(m2) Mach P(Pa) rho(kg/m3) T(K) c(m2/s) u(m/s)'
      iwarn = 0
      do ii = 1, nptx
        if (xsec(ii).lt.0.d0) then
          machi  = 0.1d0
          aire   = sec(ii)/secstar1
          call newton(nitmx,iwarn,gama,machi,aire,machtmp)
          call pres_isent(gama,machtmp,prestmp)
          call rho_isent(gama,machtmp,rhotmp)
          call temp_isent(gama,machtmp,temptmp)
          call sound_velocity(gama,p0left/prestmp,rho0left/rhotmp,soundtmp)
          mach(ii) = machtmp
          pres(ii) = p0left/prestmp
          rho(ii)  = rho0left/rhotmp
          temp(ii) = t0left/temptmp
          sound(ii) = soundtmp
          vel(ii) = machtmp*soundtmp
          write(impout,'(8e14.5)') xsec(ii)+xmin, sec(ii), mach(ii), pres(ii), rho(ii), &
               temp(ii), sound(ii), vel(ii)

        elseif(xsec(ii).ge.0.d0.and.sec(ii).le.secshock) then
          machi  = 1.1d0
          aire   = sec(ii)/secstar1
          call newton(nitmx,iwarn,gama,machi,aire,machtmp)
          call pres_isent(gama,machtmp,prestmp)
          call rho_isent(gama,machtmp,rhotmp)
          call temp_isent(gama,machtmp,temptmp)
          call sound_velocity(gama,p0left/prestmp,rho0left/rhotmp,soundtmp)
          mach(ii) = machtmp
          pres(ii) = p0left/prestmp
          rho(ii)  = rho0left/rhotmp
          temp(ii) = t0left/temptmp
          sound(ii) = soundtmp
          vel(ii) = machtmp*soundtmp
          write(impout,'(8e14.5)') xsec(ii)+xmin, sec(ii), mach(ii), pres(ii), rho(ii), &
               temp(ii), sound(ii), vel(ii)

        else
          machi = 0.1d0
          aire   = sec(ii)/secstar2
          call newton(nitmx,iwarn,gama,machi,aire,machtmp)
          call pres_isent(gama,machtmp,prestmp)
          call rho_isent(gama,machtmp,rhotmp)
          call temp_isent(gama,machtmp,temptmp)
          call sound_velocity(gama,p0right/prestmp,rho0right/rhotmp,soundtmp)
          mach(ii) = machtmp
          pres(ii) = p0right/prestmp
          rho(ii)  = rho0right/rhotmp
          temp(ii) = t0right/temptmp
          sound(ii) = soundtmp
          vel(ii) = machtmp*soundtmp
          write(impout,'(8e14.5)') xsec(ii)+xmin, sec(ii), mach(ii), pres(ii), rho(ii), &
               temp(ii), sound(ii), vel(ii)

        endif
      enddo
      close(impout)

      ! Calculation of the mass flow rate,
      ! and writing of the inlet/Exhaust conditions
      massflow = rho(1)*mach(1)*sound(1)*sec(1)

      write(nfecra,*) '--------------------------------------------------------------'
      write(nfecra,*) '  Mass flow rate (kg/s) : ', massflow
      write(nfecra,*) '--------------------------------------------------------------'
      write(nfecra,*) '  Inlet conditions'
      write(nfecra,*) '  ui   (m/s)   : ', vel(1)
      write(nfecra,*) '  Mi           : ', mach(1)
      write(nfecra,*) '  pi   (Pa)    : ', pres(1)
      write(nfecra,*) '  rhoi (kg/m3) : ', rho(1)
      write(nfecra,*) '  Ti   (K)     : ', temp(1)
      write(nfecra,*) '--------------------------------------------------------------'
      write(nfecra,*) '  Exhaust conditions'
      write(nfecra,*) '  ue   (m/s)   : ', vel(nptx)
      write(nfecra,*) '  Me           : ', mach(nptx)
      write(nfecra,*) '  pe   (Pa)    : ', pres(nptx)
      write(nfecra,*) '  rhoe (kg/m3) : ', rho(nptx)
      write(nfecra,*) '  Te   (K)     : ', temp(nptx)
      write(nfecra,*) '---------------------------------------------------------------'
      write(nfecra,*) 'Solution profiles are given in nozzle_profile.dat in this order:'
      write(nfecra,*) '# x(m) S(m2) Mach P(Pa) rho(kg/m3) T(K) c(m2/s) u(m/s)'

    endif
  endif

  !===============================================================================

  deallocate(xsec)
  deallocate(sec)
  deallocate(mach)
  deallocate(pres)
  deallocate(rho)
  deallocate(temp)
  deallocate(sound)
  deallocate(vel)

endif

!--------
! Formats
!--------

!----
! End
!----


return
end subroutine cs_f_user_extra_operations
